import dateparser
from datetime import datetime, timezone

import requests
from typing import Dict, Union, Tuple, TypeVar, Callable, List, Type, Sequence

import math
from shapely import wkt, errors

number = TypeVar('number', int, float)


def parse_string(value: str) -> str:
    """
    Base string validator. Maybe silly, but we can also ensure any constraints needed in the future.
    :param value: The string to validate
    :return: The validated string, with any required modifications
    """
    # Convert to string first, so length is checked against only str types:
    try:
        value = f'{value}'
    except ValueError as exc:  # If this happens, printing v's value would fail too...
        raise ValueError(f"Invalid string: Can't cast type '{type(value)}' to string.") from exc
    if len(value) == 0:
        raise ValueError('Invalid string: Empty.')
    return value


def parse_float(value: float) -> float:
    """
    Base float validator. Ensures values like Inf are not allowed even though they are valid floats.
    :param value: The float to validate
    :return: The validated float
    """
    try:
        value = float(value)
    except ValueError as exc:
        raise ValueError(f'Invalid float: {value}') from exc
    if math.isinf(value) or math.isnan(value):
        raise ValueError(f'Float values must be finite: got {value}')
    return value

def parse_int(value: int) -> int:
    """
    Base int validator.
    :param value: The int to validate
    :return: The validated int
    """
    try:
        value = int(value)
    except ValueError as exc:
        raise ValueError(f'Invalid int: {value}') from exc
    return value

def parse_date(value: Union[str, datetime]) -> Union[datetime, str]:
    """
    Base date validator
    :param value: String or datetime object to be validated
    :return: String passed in, if it can successfully convert to Datetime.
    (Need to keep strings like "today" w/out converting them, but throw on "asdf")
    """
    if isinstance(value, datetime):
        return _to_utc(value)

    date = dateparser.parse(str(value))
    if date is None:
        raise ValueError(f"Invalid date: '{value}'.")

    return _to_utc(date).strftime('%Y-%m-%dT%H:%M:%SZ')


def _to_utc(date: datetime):
    if date.tzinfo is None:
        date = date.replace(tzinfo=timezone.utc)
    return date


def parse_range(
    value: Tuple[number, number], h: Callable[[number], number]
) -> Tuple[number, number]:
    """
    Base range validator. For our purposes, a range is a tuple
    with exactly two numeric elements (a, b), requiring a <= b.

    Parameters
    ----------
    value: The range to be validated. Examples: (3, 5), (1.1, 12.3)
    h: The validator function to apply to each individual value

    Returns
    ----------
    Validated tuple representing the range
    """
    if isinstance(value, tuple):
        if len(value) < 2:
            raise ValueError(f'Not enough values in min/max tuple: {value}')
        if len(value) > 2:
            raise ValueError(f'Too many values in min/max tuple: {value}')
        value = (h(value[0]), h(value[1]))
        if math.isinf(value[0]) or math.isnan(value[0]):
            raise ValueError(
                f'Expected finite numeric min in min/max tuple, got {value[0]}: {value}'
            )
        if math.isinf(value[1]) or math.isnan(value[1]):
            raise ValueError(
                f'Expected finite numeric max in min/max tuple, got {value[1]}: {value}'
            )
        if value[0] > value[1]:
            raise ValueError(
                f'Min must be less than max when using min/max tuples to search: {value}'
            )
        return value
    raise ValueError(f'Invalid range. Expected 2-value numeric tuple, got {type(value)}: {value}')


# Parse and validate a date range: "1991-10-01T00:00:00Z,1991-10-02T00:00:00Z"
def parse_date_range(
    value: Tuple[Union[str, datetime], Union[str, datetime]],
) -> Tuple[datetime, datetime]:
    return parse_range(value, parse_date)


# Parse and validate an integer range: "3-5"
def parse_int_range(value: Tuple[int, int]) -> Tuple[int, int]:
    return parse_range(value, int)


# Parse and validate a float range: "1.1-12.3"
def parse_float_range(value: Tuple[float, float]) -> Tuple[float, float]:
    return parse_range(value, float)


# Parse and validate an iterable of values, using h() to validate each value:
# "a,b,c", "1,2,3", "1.1,2.3"
def parse_list(value: Sequence, h) -> List:
    if not isinstance(value, Sequence) or isinstance(value, str):
        value = [value]
    try:
        return [h(a) for a in value]
    except ValueError as exc:
        raise ValueError(f'Invalid {h.__name__} list: {exc}') from exc


def parse_cmr_keywords_list(value: Sequence[Union[Dict, Sequence]]):
    if not isinstance(value, Sequence) or (
        len(value) == 2 and isinstance(value[0], str)
    ):  # in case we're passed single key value pair as sequence
        value = [value]

    for idx, item in enumerate(value):
        if not isinstance(item, tuple) and not isinstance(item, Sequence):
            raise ValueError(
                f'Expected item in cmr_keywords list index {idx} to be tuple pair, '
                f'got value {item} of type {type(item)}'
            )
        if len(item) != 2:
            raise ValueError(
                f'Expected item in cmr_keywords list index {idx} to be of length 2, '
                f'got value {item} of length {len(item)}'
            )

        search_key, search_value = item
        if not isinstance(search_key, str) or not isinstance(search_value, str):
            raise ValueError(
                f'Expected tuple pair of types: '
                f'"{type(str)}, {type(str)}" in cmr_keywords at index {idx}, '
                f'got value "{str(item)}" '
                f'of types: "{type(search_key)}, {type(search_value)}"'
            )

    return value


# Parse and validate an iterable of strings: "foo,bar,baz"
def parse_string_list(value: Sequence[str]) -> List[str]:
    return parse_list(value, parse_string)


# Parse and validate an iterable of integers: "1,2,3"
def parse_int_list(value: Sequence[int]) -> List[int]:
    return parse_list(value, int)


# Parse and validate an iterable of floats: "1.1,2.3,4.5"
def parse_float_list(value: Sequence[float]) -> List[float]:
    return parse_list(value, float)


def parse_number_or_range(value: Union[List, Tuple[number, number], range], h):
    try:
        if isinstance(value, tuple):
            return parse_range(value, h)
        if isinstance(value, range):
            if value.step == 1:
                return [value.start, value.stop]

        return h(value)

    except ValueError as exc:
        raise ValueError(f'Invalid {h.__name__} or range: {exc}') from exc


# Parse and validate an iterable of numbers or number ranges, using h() to validate each value:
# "1,2,3-5", "1.1,1.4,5.1-6.7"
def parse_number_or_range_list(value: Sequence, h) -> List:
    if not isinstance(value, Sequence) or isinstance(value, range):
        value = [value]

    return [parse_number_or_range(x, h) for x in value]


# Parse and validate an iterable of integers or integer ranges: "1,2,3-5"
def parse_int_or_range_list(value: Sequence) -> List:
    return parse_number_or_range_list(value, int)


# Parse and validate an iterable of float or float ranges: "1.0,2.0,3.0-5.0"
def parse_float_or_range_list(value: Sequence) -> List:
    return parse_number_or_range_list(value, parse_float)


# Parse and validate a coordinate list
def parse_coord_list(value: Sequence[float]) -> List[float]:
    if not isinstance(value, Sequence):
        raise ValueError(f'Invalid coord list list: Must pass in an iterable. Got {type(value)}.')
    for coord in value:
        try:
            float(coord)
        except ValueError as exc:
            raise ValueError(f'Invalid coordinate: {coord}') from exc
    if len(value) % 2 != 0:
        raise ValueError(f'Invalid coordinate list, odd number of values provided: {value}')
    return value


# Parse and validate a bbox coordinate list
def parse_bbox_list(value: Sequence[float]) -> List[float]:
    try:
        # This also makes sure v is an iterable:
        value = parse_coord_list(value)
    except ValueError as exc:
        raise ValueError(f'Invalid bbox: {exc}') from exc
    if len(value) != 4:
        raise ValueError(f'Invalid bbox, must be 4 values: {value}')
    return value


# Parse and validate a point coordinate list
def parse_point_list(value: Sequence[float]) -> List[float]:
    try:
        # This also makes sure v is an iterable:
        value = parse_coord_list(value)
    except ValueError as exc:
        raise ValueError(f'Invalid point: {exc}') from exc
    if len(value) != 2:
        raise ValueError(f'Invalid point, must be 2 values: {value}')
    return value


# Parse a WKT and convert it to a coordinate string
def parse_wkt(value: str) -> str:
    try:
        value = wkt.loads(value)
    except errors.WKTReadingError as exc:
        raise ValueError(f'Invalid wkt: {exc}') from exc
    return wkt.dumps(value)


# Parse a CMR circle:
#       [longitude, latitude, radius(meters)]
def parse_circle(value: List[float]) -> List[float]:
    value = parse_float_list(value)
    if len(value) != 3:
        raise ValueError(f'Invalid circle, must be 3 values (lat, long, radius). Got: {value}')
    return value


# Parse a CMR linestring:
#       [longitude, latitude, longitude, latitude, ...]
def parse_linestring(value: List[float]) -> List[float]:
    value = parse_float_list(value)
    if len(value) % 2 != 0:
        raise ValueError(
            f'Invalid linestring, must be values of format (lat, long, lat, long, ...). Got: {value}'
        )
    return value


def parse_point(value: List[float]) -> List[float]:
    value = parse_float_list(value)
    if len(value) != 2:
        raise ValueError(f'Invalid point, must be values of format (lat, long). Got: {value}')
    return value


# Parse and validate a coordinate string
def parse_coord_string(value: List):
    value = parse_float_list(value)
    if len(value) % 2 != 0:
        raise ValueError(
            f'Invalid coordinate string, must be values of format (lat, long, lat, long, ...). Got: {value}'
        )
    return value

def parse_bbox(value: List[float]) -> List[float]:
    value = parse_float_list(value)
    if len(value) != 4:
        raise ValueError(
            f'Invalid bbox string, must be values of format (min_lat, min_lon, max_lat, max_lon). Got: {value}'
        )

    if value[0] > 180 or value[2] > 180:
        value = [
            (x + 180) % 360 - 180 if idx % 2 == 0 and abs(x) > 180 else x
            for idx, x in enumerate(value)
        ]

        bottom_left = [coord for coord in value[:2]]
        top_right = [coord for coord in value[2:]]

        value = [*bottom_left, *top_right]
    return value

# Take "requests.Session", or anything that subclasses it:
def parse_session(session: Type[requests.Session]):
    if issubclass(type(session), requests.Session):
        return session
    else:
        raise ValueError(
            'Invalid Session: expected ASFSession or a requests.Session subclass. '
            f'Got {type(session)}'
        )



from datetime import date
import pickle
from pathlib import Path

from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search import SBASNetwork

import pytest

@pytest.fixture
def reference():
    path = Path(__file__).parent / "data/reference_scene_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def pair():
    path = Path(__file__).parent / "data/pair_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def stack_results():
    path = Path(__file__).parent / "data/stack_search_0.pkl"

    with open(path, "rb") as f:
        # trusted local test fixture
        return pickle.load(f) # nosec B301
    
@pytest.fixture
def sbas_network(stack_results):
    return SBASNetwork.from_search_results(
        stack_results,
        season=(1, 176),
        perpendicular_baseline=100, 
        inseason_temporal_baseline=24,
        bridge_target_date='3-1',
        bridge_year_threshold=2)
    
def test_sbas_network_from_ref_scene(mocker, reference, stack_results):
    """
    Create an SBASNetwork from a geographic reference scene
    """
    mock_stack = mocker.patch.object(
        reference,
        "stack",
        return_value=stack_results,
    )

    # Create an SBASNetwork and confrm expected size of its full_stack and subset_stack
    sbas = SBASNetwork(
        reference,
        start_date="2020-01-01",
        end_date="2025-10-02",
        season=(1, 176),
        perpendicular_baseline=100, 
        inseason_temporal_baseline=24,
        bridge_target_date='3-1',
        bridge_year_threshold=2)

    assert len(sbas.full_stack) == 1730
    assert len(sbas.subset_stack) == 176
    assert len(sbas.connected_substacks) == 3
    assert len(max(sbas.connected_substacks, key=lambda s: len(s))) == 161

def test_sbasnetwork_from_search_results(sbas_network):
    """
    Create an SBASNetwork from ASFProduct.stack search results with the 
    SBASNetwork.from_search_results alternate class method constructor
    """
    assert len(sbas_network.full_stack) == 1730
    assert len(sbas_network.subset_stack) == 176
    assert len(max(sbas_network.connected_substacks, key=lambda s: len(s))) == 161

def test_disallow_missing_state_vectors(stack_results):
    """
    Test the optional allow_missing_state_vectors argument set to False (default). 
    """
    stack_results[2].baseline["stateVectors"]["positions"]["prePositionTime"] = None
    missing_state_vectors_not_allowed_sbas = SBASNetwork.from_search_results(
        stack_results,
        season=(1, 176),
        perpendicular_baseline=100, 
        inseason_temporal_baseline=24,
        bridge_target_date='3-1',
        bridge_year_threshold=2)
    assert len(missing_state_vectors_not_allowed_sbas.full_stack) == 1698
    assert len([p for p in missing_state_vectors_not_allowed_sbas.full_stack if p.perpendicular_baseline is None]) == 0

def test_disallow_missing_state_vectors(stack_results):
    """
    Test the optional allow_missing_state_vectors argument set to True. 
    """
    stack_results[2].baseline["stateVectors"]["positions"]["prePositionTime"] = None
    missing_state_vectors_allowed_sbas = SBASNetwork.from_search_results(
        stack_results,
        season=(1, 176),
        perpendicular_baseline=100, 
        inseason_temporal_baseline=24,
        bridge_target_date='3-1',
        bridge_year_threshold=2,
        allow_missing_state_vectors=True)
    assert len(missing_state_vectors_allowed_sbas.full_stack) == 1730
    assert len([p for p in missing_state_vectors_allowed_sbas.full_stack if p.perpendicular_baseline is None]) == 32

def test_add_pair(sbas_network):
    removed_pair_0 = sbas_network.get_pair_from_dates(sbas_network.remove_list, date(2023,1,17), date(2024,6,16))
    sbas_network.add_pairs(removed_pair_0)
    assert len(sbas_network.subset_stack) == 177
    assert len(sbas_network.connected_substacks) == 2
    assert len(max(sbas_network.connected_substacks, key=lambda s: len(s))) == 165

def test_add_pairs(sbas_network):
    removed_pair_0 = sbas_network.get_pair_from_dates(sbas_network.remove_list, date(2023,1,17), date(2024,6,16))
    removed_pair_1 = sbas_network.get_pair_from_dates(sbas_network.remove_list, date(2023,1,5), date(2024,2,5))
    sbas_network.add_pairs([removed_pair_0, removed_pair_1])
    assert len(sbas_network.subset_stack) == 178
    assert len(sbas_network.connected_substacks) == 2
    assert len(max(sbas_network.connected_substacks, key=lambda s: len(s))) == 166

def test_remove_pairs(sbas_network):
    pair1_to_remove = sbas_network.get_pair_from_dates(sbas_network.subset_stack, date(2025,1,6), date(2025,1,18))
    pair2_to_remove = sbas_network.get_pair_from_dates(sbas_network.subset_stack, date(2025,1,6), date(2025,1,30))
    sbas_network.remove_pairs([pair1_to_remove, pair2_to_remove])
    assert len(sbas_network.subset_stack) == 174
    assert len(max(sbas_network.connected_substacks, key=lambda s: len(s))) == 159
    assert len(sbas_network.remove_list) == 1556

def add_new_pair(sbas_network, pair):
    sbas_network.add_pairs(pair)
    assert len(sbas_network.full_stack) == 1731
    assert len(sbas_network.subset_stack) == 176
    assert len(max(sbas_network.connected_substacks, key=lambda s: len(s))) == 164



import logging
from typing import Union, Tuple, List
from shapely import wkt
from shapely.geometry.base import BaseGeometry
from shapely.geometry import (
    Polygon,
    MultiPolygon,
    Point,
    LineString,
    GeometryCollection,
)
from shapely.geometry.collection import BaseMultipartGeometry
from shapely.ops import transform, orient, unary_union
from .RepairEntry import RepairEntry

from asf_search.exceptions import ASFWKTError


def validate_wkt(
    aoi: Union[str, BaseGeometry],
) -> Tuple[BaseGeometry, BaseGeometry, List[RepairEntry]]:
    """
    Param aoi: the WKT string or Shapely Geometry to validate and prepare for the CMR query
    Validates the given area of interest, and returns a validated and simplified WKT string
    returns: The input AOI's CMR ready WKT string
    """
    if isinstance(aoi, str):
        aoi_shape = wkt.loads(aoi)
    else:
        aoi_shape = wkt.loads(aoi.wkt)

    if not aoi_shape.is_valid:
        aoi_shape = _search_wkt_prep(aoi_shape)

        if not aoi_shape.is_valid and not isinstance(aoi_shape, MultiPolygon):
            if isinstance(aoi_shape, Polygon):
                if not aoi_shape.exterior.is_simple:
                    raise ASFWKTError(
                        f'WKT string: "{aoi_shape.wkt}" is a self intersecting polygon'
                    )

            raise ASFWKTError(f'WKT string: "{aoi_shape.wkt}" is not a valid WKT string')

    if aoi_shape.is_empty:
        raise ASFWKTError(f'WKT string: "{aoi_shape.wkt}" empty WKT is not a valid AOI')

    wrapped, unwrapped, reports = _simplify_geometry(aoi_shape)

    return wrapped, unwrapped, [report for report in reports if report is not None]


def _search_wkt_prep(shape: BaseGeometry):
    if isinstance(shape, MultiPolygon):
        output = []
        for geom in shape.geoms:
            output.append(orient(Polygon(geom.exterior)))

        return MultiPolygon(output)

    if isinstance(shape, Polygon):
        return orient(Polygon(shape.exterior), sign=1.0)


def _simplify_geometry(
    geometry: BaseGeometry,
) -> Tuple[BaseGeometry, BaseGeometry, List[RepairEntry]]:
    """
    param geometry: AOI Shapely Geometry to be prepped for CMR
    prepares geometry for CMR by:
        1. Flattening any nested multi-part geometry into single collection
        2. clamping latitude +/-90, unwrapping longitude +/-180,
            removing coordinate dimensions higher than 2 (lon,lat)
        3. Merging any overlapping shapes
        4. convex-hulling the remainder into a single shape
        4. simplifing until the shape has <= 300 points, with no point closer than 0.00001
        5. Orienting vertices in counter-clockwise winding order
    returns: geometry prepped for CMR
    """
    flattened = _flatten_multipart_geometry(geometry)

    merged, merge_report = _merge_overlapping_geometry(flattened)
    convex, convex_report = _get_convex_hull(merged)
    simplified, simplified_report = _simplify_aoi(convex)
    reoriented, reorientation_report = _counter_clockwise_reorientation(simplified)
    wrapped, unwrapped, clamp_report = _get_clamped_and_wrapped_geometry(reoriented)

    dimension_report = (
        RepairEntry(
            report_type="'type': 'EXTRA_DIMENSION'",
            report="'report': Only 2-Dimensional area of interests are supported (lon/lat), "
            'higher dimension coordinates will be ignored',
        )
        if geometry.has_z
        else None
    )

    if convex_report is not None:
        merge_report = None

    repair_reports = [
        dimension_report,
        merge_report,
        convex_report,
        *clamp_report,
        *simplified_report,
        reorientation_report,
    ]
    for report in repair_reports:
        if report is not None:
            logging.info(f'{report}')

    validated_wrapped = transform(lambda x, y, z=None: tuple([round(x, 14), round(y, 14)]), wrapped)
    validated_unwrapped = transform(
        lambda x, y, z=None: tuple([round(x, 14), round(y, 14)]), unwrapped
    )
    return validated_wrapped, validated_unwrapped, repair_reports


def _flatten_multipart_geometry(unflattened_geometry: BaseGeometry) -> BaseGeometry:
    """
    Recursively flattens nested geometric collections,
    guarantees geometric collections have a depth equal to 1.
    Also ignores any empty shapes in multipart geometry
    """

    def _recurse_nested_geometry(geometry: BaseGeometry) -> List[BaseGeometry]:
        output = []

        if isinstance(geometry, BaseMultipartGeometry):
            for geom in geometry.geoms:
                output.extend(_recurse_nested_geometry(geom))
        elif not geometry.is_empty:
            if isinstance(geometry, Polygon):
                return [Polygon(geometry.exterior)]
            return [geometry]

        return output

    flattened = _recurse_nested_geometry(unflattened_geometry)

    return flattened[0] if len(flattened) == 1 else GeometryCollection(flattened)


def _merge_overlapping_geometry(
    geometry: BaseGeometry,
) -> Tuple[BaseGeometry, RepairEntry]:
    """
    parameter geometry: geometry to merge
    Performs a unary union overlapping operation of the input geometry,
    ensuring geometric collections (multipolygon, multipartgeometry, etc)
    are simplied as much as possible before the convex-hull step
    output: merged-overlapping geometry
    """
    merge_report = None

    if isinstance(geometry, BaseMultipartGeometry):
        original_amount = len(geometry.geoms)
        if original_amount == 1:
            return geometry, merge_report

        merged = unary_union(geometry)

        # if there were non-overlapping shapes
        if isinstance(merged, BaseMultipartGeometry):
            unique_shapes = len(merged.geoms)
            merged = orient(
                unary_union(GeometryCollection([geom.convex_hull for geom in merged.geoms]))
            )
            if isinstance(merged, BaseMultipartGeometry):
                if unique_shapes != len(merged.geoms):
                    merge_report = RepairEntry(
                        "'type': 'OVERLAP_MERGE'",
                        f"'report': {unique_shapes - len(merged.geoms)} "
                        'non-overlapping shapes merged by their convex-hulls',
                    )
            else:
                merge_report = RepairEntry(
                    "'type': 'OVERLAP_MERGE'",
                    f"'report': {unique_shapes} non-overlapping shapes merged by their convex-hulls",  # noqa F401
                )
        else:
            merge_report = RepairEntry(
                "'type': 'OVERLAP_MERGE'",
                f"'report': Overlapping {original_amount} shapes merged into one",
            )

        return merged, merge_report

    return geometry, merge_report


def _counter_clockwise_reorientation(geometry: Union[Point, LineString, Polygon]):
    """
    param geometry: Shapely geometry to re-orient
    Ensures the geometry coordinates are wound counter-clockwise
    output: counter-clockwise oriented geometry
    """
    reoriented_report = RepairEntry("'type': 'REVERSE'", "'report': Reversed polygon winding order")
    reoriented = orient(geometry)

    if isinstance(geometry, Polygon):
        # if the vertice ordering has changed
        if reoriented.exterior.is_ccw != geometry.exterior.is_ccw:
            return reoriented, reoriented_report

    return reoriented, None


def _get_clamped_and_wrapped_geometry(
    shape: BaseGeometry,
) -> Tuple[BaseGeometry, BaseGeometry, List[RepairEntry]]:
    """
    param geometry: Shapely geometry to clamp
    Clamps geometry to +/-90 latitude and wraps longitude +/-180
    output: clamped shapely geometry
    """
    coords_clamped = 0
    coords_wrapped = 0

    def _clamp_lat(x, y, z=None):
        clamped = _clamp(y)

        if clamped != y:
            nonlocal coords_clamped
            coords_clamped += 1

        return tuple([x, clamped])

    def _wrap_lon(x, y, z=None):
        wrapped = x
        if abs(x) > 180:
            wrapped = (x + 180) % 360 - 180

        if wrapped != x:
            nonlocal coords_wrapped
            coords_wrapped += 1

        return tuple([wrapped, y])

    def _unwrap_lon(x, y, z=None):
        unwrapped = x if x >= 0 else x + 360  # This undoes wrapping

        return tuple([unwrapped, y])

    clamped_lat = transform(_clamp_lat, shape)

    wrapped = transform(_wrap_lon, clamped_lat)

    if wrapped.bounds[2] - wrapped.bounds[0] > 180:
        unwrapped = transform(_unwrap_lon, wrapped)
    else:
        unwrapped = wrapped

    clampRepairReport = None
    wrapRepairReport = None

    if coords_clamped > 0:
        clampRepairReport = RepairEntry(
            "'type': 'CLAMP'",
            f"'report': 'Clamped {coords_clamped} value(s) to +/-90 latitude'",
        )
    if coords_wrapped > 0:
        wrapRepairReport = RepairEntry(
            "'type': 'WRAP'",
            f"'report': 'Wrapped {coords_wrapped} value(s) to +/-180 longitude'",
        )

    return (wrapped, unwrapped, [clampRepairReport, wrapRepairReport])


def _get_convex_hull(geometry: BaseGeometry) -> Tuple[BaseGeometry, RepairEntry]:
    """
    param geometry: geometry to perform possible convex hull operation on
    If the given geometry is a collection of geometries,
    creates a convex-hull encompassing said geometry
    output: convex hull of multi-part geometry, or the original single-shaped geometry
    """
    if geometry.geom_type not in [
        'MultiPoint',
        'MultiLineString',
        'MultiPolygon',
        'GeometryCollection',
    ]:
        return geometry, None

    possible_repair = RepairEntry(
        "'type': 'CONVEX_HULL_INDIVIDUAL'",
        "'report': 'Unconnected shapes: Convex-hulled each INDIVIDUAL shape to merge them together.'",  # noqa F401
    )
    return geometry.convex_hull, possible_repair


def _simplify_aoi(
    shape: Union[Polygon, LineString, Point],
    threshold: float = 0.0004,
    max_depth: int = 20,
) -> Tuple[Union[Polygon, LineString, Point], List[RepairEntry]]:
    """
    param shape: Shapely geometry to simplify
    param threshold: point proximity threshold to merge nearby points of geometry with
    param max_depth: the max amount of iterations to simplify, defaults to 20
    Simplifies geometry with increasing threshold, until there are no more than 300 points
    output: simplified geometry
    """
    repairs = []

    if shape.geom_type == 'Point':
        return shape, repairs

    # Check for very small shapes and collapse accordingly
    mbr_width = shape.bounds[2] - shape.bounds[0]
    mbr_height = shape.bounds[3] - shape.bounds[1]
    # If both pass, it's a tiny box. Turn it to a point
    if mbr_width <= threshold and mbr_height <= threshold:
        simplified = shape.centroid
        repair = RepairEntry(
            "'type': 'GEOMETRY_SIMPLIFICATION'",
            "'report': 'Shape Collapsed to Point: "
            f'shape of {_get_shape_coords_len(shape)} '
            f'simplified to {_get_shape_coords_len(simplified)} '
            f"with proximity threshold of {threshold}'",
        )
        return simplified, [*repairs, repair]
    # If it's a single line segment, it's already as simple as can be. Don't do anything
    elif shape.geom_type == 'LineString' and len(shape.coords) == 2:
        return shape, repairs
    # Else, check if it's slim enough to become a linestring:
    elif mbr_width <= threshold:
        lon = (shape.bounds[2] - shape.bounds[0]) / 2 + shape.bounds[0]
        simplified = LineString([(lon, shape.bounds[1]), (lon, shape.bounds[3])])
        repair = RepairEntry(
            "'type': 'GEOMETRY_SIMPLIFICATION'",
            f"'report': 'Shape Collapsed to Vertical Line: shape of {_get_shape_coords_len(shape)} "
            f'simplified to {_get_shape_coords_len(simplified)} '
            f"with proximity threshold of {threshold}'",
        )
        return simplified, [*repairs, repair]
    elif mbr_height <= threshold:
        lat = (shape.bounds[3] - shape.bounds[1]) / 2 + shape.bounds[1]
        simplified = LineString([(shape.bounds[0], lat), (shape.bounds[2], lat)])
        repair = RepairEntry(
            "'type': 'GEOMETRY_SIMPLIFICATION'",
            "'report': 'Shape Collapsed to Horizontal Line: "
            f'shape of {_get_shape_coords_len(shape)} simplified '
            f"to {_get_shape_coords_len(simplified)} with proximity threshold of {threshold}'",
        )
        return simplified, [*repairs, repair]
    
    # Keep taking away points until it's under 300:
    for simplify_level in range(0, max_depth):
        simplifed = shape.simplify(tolerance=threshold * (1.5**simplify_level))

        coords_length = _get_shape_coords_len(simplifed)
        if _get_shape_coords_len(shape) != coords_length:
            repairs.append(
                RepairEntry(
                    "'type': 'GEOMETRY_SIMPLIFICATION'",
                    f"'report': 'Shape Simplified: shape of {_get_shape_coords_len(shape)} "
                    f"simplified to {coords_length} with proximity threshold of {threshold}'",
                )
            )

        if coords_length <= 300:
            return simplifed, repairs

    raise ASFWKTError(f'Failed to simplify wkt string: {shape.wkt}')


def _clamp(num):
    """Clamps value between -90 and 90"""
    return max(-90, min(90, num))


def _get_shape_coords_len(geometry: BaseGeometry):
    return len(_get_shape_coords(geometry))


def _get_shape_coords(geometry: BaseGeometry):
    """Returns flattened coordinates of input Shapely geometry"""
    if geometry.geom_type == 'Polygon':
        return list(geometry.exterior.coords[:-1])

    if geometry.geom_type == 'LineString':
        return list(geometry.coords)

    if geometry.geom_type == 'Point':
        return list(geometry.coords)

    output = []

    for geom in geometry.geoms:
        coords = _get_shape_coords(geom)
        output = [*output, *coords]

    return output



from typing import Dict
import requests
import json

import asf_search.constants


def health(host: str = None) -> Dict:
    """
    Checks basic connectivity to and health of the ASF SearchAPI.

    Parameters
    ----------
    param host:
        SearchAPI host, defaults to Production SearchAPI.
        This option is intended for dev/test purposes.

    Returns
    -------
    Current configuration and status of subsystems as a dict
    """

    if host is None:
        host = asf_search.INTERNAL.CMR_HOST
    return json.loads(requests.get(f'https://{host}{asf_search.INTERNAL.CMR_HEALTH_PATH}').text)



from asf_search.constants import PRODUCT_TYPE
from asf_search.search import search
from asf_search import ASFSearchOptions
from copy import copy


def get_nisar_orbit_ephemeras(opts: ASFSearchOptions | None = None):
    """Returns a dictionary of the latest NISAR orbit ephemera products (`POE`, `MOE`, `NOE`, and `FOE`).
    Additional options may be specified to pass to search.

    For more information refer to the NISAR Data User Guide:
    https://nisar-docs.asf.alaska.edu/orbit-ephemeris/#tbl-nisar-orbit-ephemeris-characteristics
    """
    opts = ASFSearchOptions() if opts is None else copy(opts)

    opts.shortName = 'NISAR_OE'
    opts.maxResults = 1

    orbits = {}
    for processingLevel in [PRODUCT_TYPE.POE, PRODUCT_TYPE.MOE, PRODUCT_TYPE.NOE, PRODUCT_TYPE.FOE]:
        orbits[processingLevel] = search(opts=opts, processingLevel=processingLevel)[0]

    return orbits



from asf_search import Products, search, ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults
import json
import pytest
from asf_search.constants import PRODUCT_TYPE

def run_test_ASFSubproduct(scene_names: list[str], expected_subclass: str, opts: ASFSearchOptions):
    scenes = search(granule_list=scene_names, opts=opts)

    assert sorted([scene.properties['fileID'] for scene in scenes]) == sorted(scene_names)

    for scene in scenes:
        assert expected_subclass.upper() == scene.__class__.__name__ .upper(), f'Expected scene "{scene.properties["fileID"]}" to be of ASFProduct subclass {expected_subclass}. Got {scene.__class__.__name__}'
        if isinstance(scene, Products.OPERAS1Product):
            _test_OPERAS1Product(scene)
        if isinstance(scene, Products.S1BurstProduct):
            _test_S1BurstProduct(scene)
        if isinstance(scene, Products.SEASATProduct):
            _test_SEASATProduct(scene)
        if isinstance(scene, Products.TROPOProduct):
            _test_TROPOProduct(scene)

    for output_format in ['geojson', 'json', 'jsonlite', 'jsonlite2', 'csv', 'metalink', 'kml']:
        try:
            _get_output(scenes, output_format)
        except BaseException as exc:
            pytest.fail(f'Failed to serialized scenes {[scene.properties["fileID"] for scene in scenes]} as output format {output_format}. Original exception: {str(exc)}')

def _test_OPERAS1Product(scene: Products.OPERAS1Product):
    processing_level = scene.properties['processingLevel']

    if processing_level in ['RTC', 'RTC-STATIC']:
        _check_properties_set(scene.properties, ['bistaticDelayCorrection'])

        if processing_level == 'RTC':
            _check_properties_set(scene.properties,['noiseCorrection', 'postProcessingFilter'])

    elif processing_level == 'DISP-S1':
        _check_properties_set(scene.properties, [
            'frameNumber', 'OperaDispStackID', 'zarrUri', 'zarrStackUri',
        ])
    
    if processing_level == 'TROPO-ZENITH':
        assert scene.properties['centerLat'] is None
        assert scene.properties['centerLon'] is None
    
def _test_SEASATProduct(scene: Products.SEASATProduct):
    assert isinstance(scene.properties['md5sum'], dict)
    assert isinstance(scene.properties['bytes'], dict)
    
    bytes_entries = scene.properties['bytes'].keys()
    _check_properties_set(scene.properties['md5sum'], bytes_entries)

def _test_TROPOProduct(scene: Products.TROPOProduct):
    assert scene.properties['processingLevel'] in [PRODUCT_TYPE.TROPO_ZENITH, PRODUCT_TYPE.ECMWF_TROPO]
    assert isinstance(scene.properties['bytes'], dict)
    
    bytes_entries = scene.properties['bytes'].keys()
    _check_properties_set(scene.properties['md5sum'], bytes_entries)

def _test_S1BurstProduct(scene: Products.S1BurstProduct):
    burst_properties = [
        "absoluteBurstID",
        "relativeBurstID",
        "fullBurstID",
        "burstIndex",
        "samplesPerBurst",
        "subswath",
        "azimuthTime",
        "azimuthAnxTime",
    ]

    _check_properties_set(scene.properties['burst'], burst_properties)


def _check_properties_set(properties: dict, properties_list: list[str]):
    for prop in properties_list:
        assert properties[prop] is not None

def _get_output(scenes: ASFSearchResults, output_format: str):
    match output_format.lower():
        case 'geojson':
            return scenes.geojson()
        case 'json':
            return json.loads(''.join(scenes.json()))
        case 'jsonlite':
            return json.loads(''.join(scenes.jsonlite()))
        case 'jsonlite2':
            return json.loads(''.join(scenes.jsonlite2()))
        case 'csv':
            return ''.join(scenes.csv())
        case 'metalink':
            return ''.join(scenes.metalink())
        case 'kml':
            return scenes.kml()




from typing import Dict, Tuple
from asf_search import ASFSearchOptions, ASFSession
from asf_search.CMR.translate import try_parse_date, try_parse_int
from asf_search.Products import S1Product


class OPERAS1Product(S1Product):
    """
    ASF Dataset Documentation Page: https://asf.alaska.edu/datasets/daac/opera/
    """

    _base_properties = {
        **S1Product._base_properties,
        'centerLat': {'path': []},  # Opera products lacks these fields
        'centerLon': {'path': []},
        'frameNumber': {'path': []},
        'operaBurstID': {'path': ['AdditionalAttributes', ('Name', 'OPERA_BURST_ID'), 'Values', 0]},
        'validityStartDate': {'path': ['TemporalExtent', 'SingleDateTime'], 'cast': try_parse_date},
        'bytes': {'path': ['DataGranule', 'ArchiveAndDistributionInformation']},
        'subswath': {'path': ['AdditionalAttributes', ('Name', 'SUBSWATH_NAME'), 'Values', 0]},
        'polarization': {
            'path': ['AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values']
        },  # dual polarization is in list rather than a 'VV+VH' style format
        'tileID': {'path': ['AdditionalAttributes', ('Name', 'MGRS_TILE_ID'), 'Values', 0]},
        'productVersion': {
            'path': ['AdditionalAttributes', ('Name', 'PRODUCT_VERSION'), 'Values', 0]
        },
    }

    _subclass_concept_ids = {
        'C1257995185-ASF',
        'C1257995186-ASF',
        'C1258354200-ASF',
        'C1258354201-ASF',
        'C1259974840-ASF',
        'C1259976861-ASF',
        'C1259981910-ASF',
        'C1259982010-ASF',
        'C2777436413-ASF',
        'C2777443834-ASF',
        'C2795135174-ASF',
        'C2795135668-ASF',
        'C1260721853-ASF',
        'C1260721945-ASF',
        'C2803501097-ASF',
        'C2803501758-ASF',
        'C3294057315-ASF',
        'C1271830354-ASF',
        'C1275699124-ASF',
        'C4090131664-ASF',
    }

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

        self.baseline = None

        self.properties['beamMode'] = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'BEAM_MODE'), 'Values', 0
        )

        self.properties['additionalUrls'] = self._get_additional_urls()
        self.properties['s3Urls'] = self._get_s3_uris()

        self.properties['operaBurstID'] = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'OPERA_BURST_ID'), 'Values', 0
        )
        self.properties['bytes'] = {
            entry['Name']: {'bytes': entry['SizeInBytes'], 'format': entry['Format']}
            for entry in self.properties['bytes']
        }

        if self.properties['processingLevel'] is None:
            self.properties['processingLevel'] = self.umm_get(
                self.umm, 'AdditionalAttributes', ('Name', 'PRODUCT_TYPE'), 'Values', 0
            )

        if self.properties['processingLevel'] == 'TROPO-ZENITH':
            self.properties['centerLat'] = None
            self.properties['centerLon'] = None
        else:
            center = self.centroid()
            self.properties['centerLat'] = center.y
            self.properties['centerLon'] = center.x

        self.properties.pop('frameNumber')

        processingLevel = self.properties['processingLevel']
        if processingLevel in [
            'RTC',
            'RTC-STATIC',
        ]:
            self.properties['bistaticDelayCorrection'] = self.umm_get(
                self.umm,
                'AdditionalAttributes',
                ('Name', 'BISTATIC_DELAY_CORRECTION'),
                'Values',
                0,
            )
            if processingLevel == 'RTC':
                self.properties['noiseCorrection'] = self.umm_get(
                    self.umm,
                    'AdditionalAttributes',
                    ('Name', 'NOISE_CORRECTION'),
                    'Values',
                    0,
                )
                self.properties['postProcessingFilter'] = self.umm_get(
                    self.umm,
                    'AdditionalAttributes',
                    ('Name', 'POST_PROCESSING_FILTER'),
                    'Values',
                    0,
                )
        elif processingLevel == 'DISP-S1':
            self.properties['frameNumber'] = try_parse_int(
                self.umm_get(
                    self.umm, 'AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0
                )
            )
            self.properties['OperaDispStackID'] = self.umm_get(
                self.umm, 'AdditionalAttributes', ('Name', 'STACK_ID'), 'Values', 0
            )
            file_id = self.properties['fileID']
            product_zarr = r'.*{0}.zarr.json.gz'.format(file_id)

            self.properties['zarrUri'] = self.find_urls(
                extension='.gz', pattern=product_zarr, directAccess=True
            )[0]
            self.properties['zarrStackUri'] = self.find_urls(
                extension='.gz',
                pattern=r'.*short_wavelength_displacement.zarr.json.gz',
                directAccess=True,
            )[0]
        elif processingLevel == 'DISP-S1-STATIC':
            self.properties['frameNumber'] = try_parse_int(
                self.umm_get(
                    self.umm, 'AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0
                )
            )
            self.properties['pathNumber'] = try_parse_int(
                self.umm_get(self.umm, 'AdditionalAttributes', ('Name', 'PATH_NUMBER'), 'Values', 0)
            )
            pass

    @staticmethod
    def get_default_baseline_product_type() -> None:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return None

    def is_valid_reference(self):
        return False

    def get_stack_opts(self, opts: ASFSearchOptions = None) -> ASFSearchOptions:
        """
        Build search options that can be used to find an insar stack for this product

        :return: ASFSearchOptions describing appropriate options
        for building a stack from this product
        """
        return None

    def get_sort_keys(self) -> Tuple[str, str]:
        keys = super().get_sort_keys()

        if keys[0] == '':
            return (self._read_property('validityStartDate', ''), keys[1])

        return keys

    @staticmethod
    def _is_subclass(item: Dict) -> bool:
        # not all umm products have this field set,
        # but when it's available it's convenient for fast matching
        concept_id = item['meta'].get('collection-concept-id')
        return concept_id in OPERAS1Product._subclass_concept_ids



from asf_search import ASFProduct, ASFStackableProduct, ASFSearchResults
from typing import Tuple, List, Union
import pytz
from .calc import calculate_perpendicular_baselines

try:
    from ciso8601 import parse_datetime
except ImportError:
    from dateutil.parser import parse as parse_datetime

def get_baseline_from_stack(
    reference: ASFProduct, stack: ASFSearchResults
) -> Tuple[ASFSearchResults, List[dict]]:
    warnings = []

    if len(stack) == 0:
        raise ValueError("No products found matching stack parameters")

    stack = [
        product
        for product in stack
        if not product.properties.get("processingLevel", '').lower().startswith("metadata") and
        product.baseline is not None
    ]
    reference, stack, reference_warnings = check_reference(reference, stack)

    if reference_warnings is not None:
        warnings.append(reference_warnings)

    stack = calculate_temporal_baselines(reference, stack)

    if reference.baseline_type == ASFStackableProduct.BaselineCalcType.PRE_CALCULATED:
        stack = offset_perpendicular_baselines(reference, stack)
    else:
        stack = calculate_perpendicular_baselines(
            reference.properties["sceneName"], stack
        )

        missing_state_vectors = _count_missing_state_vectors(stack)
        if missing_state_vectors > 0:
            warnings.append(
                {
                    "MISSING STATE VECTORS":
                    f'{missing_state_vectors} scenes in stack missing State Vectors, '
                    'perpendicular baseline not calculated for these scenes'
                }
            )

    return ASFSearchResults(stack), warnings


def _count_missing_state_vectors(stack) -> int:
    return len([scene for scene in stack if scene.baseline.get("noStateVectors")])


def find_new_reference(stack: ASFSearchResults) -> Union[ASFProduct, None]:
    for product in stack:
        if product.is_valid_reference():
            return product
    return None


def check_reference(reference: ASFProduct, stack: ASFSearchResults):
    warnings = None
    if reference.properties["sceneName"] not in [
        product.properties["sceneName"] for product in stack
    ]:  # Somehow the reference we built the stack from is missing?! Just pick one
        reference = stack[0]
        warnings = [
            {
                'NEW_REFERENCE':
                'A new reference scene had to be selected in order to calculate baseline values.'
            }
        ]

    # non-s1 is_valid_reference raise an error, while we try to find a valid s1 reference
    # do we want this behaviour for pre-calc stacks?
    if not reference.is_valid_reference():
        reference = find_new_reference(stack)
        if reference is None:
            raise ValueError(
                "No valid state vectors on any scenes in stack, this is fatal"
            )

    return reference, stack, warnings


def calculate_temporal_baselines(reference: ASFProduct, stack: ASFSearchResults):
    """
    Calculates temporal baselines for a stack of products based on a reference scene
    and injects those values into the stack.

    :param reference: The reference product from which to calculate temporal baselines.
    :param stack: The stack to operate on.
    :return: None, as the operation occurs in-place on the stack provided.
    """
    reference_time = parse_datetime(reference.properties["startTime"])
    if reference_time.tzinfo is None:
        reference_time = pytz.utc.localize(reference_time)

    for secondary in stack:
        secondary_time = parse_datetime(secondary.properties["startTime"])
        if secondary_time.tzinfo is None:
            secondary_time = pytz.utc.localize(secondary_time)
        secondary.properties["temporalBaseline"] = (
            secondary_time.date() - reference_time.date()
        ).days

    return stack


def offset_perpendicular_baselines(reference: ASFProduct, stack: ASFSearchResults):
    reference_offset = float(reference.baseline["insarBaseline"])

    for product in stack:
        product.properties["perpendicularBaseline"] = round(
            float(product.baseline["insarBaseline"]) - reference_offset
        )

    return stack



import datetime
from typing import Literal, Sequence, Tuple, Union
from copy import copy
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR.subquery import build_subqueries
from asf_search.CMR import translate_opts
from asf_search.search.search_generator import get_page, preprocess_opts
from asf_search import INTERNAL


def search_count(
    absoluteOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    asfFrame: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    beamMode: Union[str, Sequence[str]] = None,
    beamSwath: Union[str, Sequence[str]] = None,
    campaign: Union[str, Sequence[str]] = None,
    maxDoppler: float = None,
    minDoppler: float = None,
    end: Union[datetime.datetime, str] = None,
    maxFaradayRotation: float = None,
    minFaradayRotation: float = None,
    flightDirection: str = None,
    flightLine: str = None,
    frame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
    granule_list: Union[str, Sequence[str]] = None,
    groupID: Union[str, Sequence[str]] = None,
    insarStackId: str = None,
    instrument: Union[str, Sequence[str]] = None,
    intersectsWith: str = None,
    lookDirection: Union[str, Sequence[str]] = None,
    offNadirAngle: Union[
        float, Tuple[float, float], Sequence[Union[float, Tuple[float, float]]]
    ] = None,
    platform: Union[str, Sequence[str]] = None,
    polarization: Union[str, Sequence[str]] = None,
    processingDate: Union[datetime.datetime, str] = None,
    processingLevel: Union[str, Sequence[str]] = None,
    product_list: Union[str, Sequence[str]] = None,
    relativeOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    season: Tuple[int, int] = None,
    start: Union[datetime.datetime, str] = None,
    ariaVersion: str = None,
    absoluteBurstID: Union[int, Sequence[int]] = None,
    relativeBurstID: Union[int, Sequence[int]] = None,
    fullBurstID: Union[str, Sequence[str]] = None,
    temporalBaselineDays: Union[str, Sequence[str]] = None,
    operaBurstID: Union[str, Sequence[str]] = None,
    frameCoverage: Literal['FULL', 'PARTIAL'] = None,
    mainBandPolarization: Union[str, Sequence[str]] = None,
    sideBandPolarization: Union[str, Sequence[str]] = None,
    rangeBandwidth: Union[str, Sequence[str]] = None,
    productionConfiguration: Union[Literal['PR', 'UR'], Sequence[Literal['PR', 'UR']]] = None,
    jointObservation: bool = None,
    dataset: Union[str, Sequence[str]] = None,
    collections: Union[str, Sequence[str]] = None,
    shortName: Union[str, Sequence[str]] = None,
    cmr_keywords: Union[Tuple[str, str], Sequence[Tuple[str, str]]] = None,
    tileID: Union[str, Sequence[str]] = None,
    maxResults: int = None,
    opts: ASFSearchOptions = None,
) -> int:
    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = ASFSearchOptions() if kwargs['opts'] is None else copy(opts)
    del kwargs['opts']

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    preprocess_opts(opts)

    url = '/'.join(s.strip('/') for s in [f'https://{opts.host}', f'{INTERNAL.CMR_GRANULE_PATH}'])

    count = 0
    for query in build_subqueries(opts):
        translated_opts = translate_opts(query)
        idx = translated_opts.index(('page_size', INTERNAL.CMR_PAGE_SIZE))
        translated_opts[idx] = ('page_size', 0)

        response = get_page(session=opts.session, url=url, translated_opts=translated_opts)
        count += response.json()['hits']
    return count



ASF_AUTH_HOST = 'cumulus.asf.alaska.edu'

CMR_HOST = 'cmr.earthdata.nasa.gov'
CMR_HOST_UAT = 'cmr.uat.earthdata.nasa.gov'
CMR_TIMEOUT = 30
CMR_FORMAT_EXT = 'umm_json'
CMR_GRANULE_PATH = f'/search/granules.{CMR_FORMAT_EXT}'
CMR_COLLECTIONS = '/search/collections'
CMR_COLLECTIONS_PATH = f'{CMR_COLLECTIONS}.{CMR_FORMAT_EXT}'
CMR_HEALTH_PATH = '/search/health'
CMR_PAGE_SIZE = 250
EDL_HOST = 'urs.earthdata.nasa.gov'
EDL_HOST_UAT = f'uat.{EDL_HOST}'

EDL_CLIENT_ID = 'BO_n7nTIlMljdvU6kRRB3g'

DEFAULT_PROVIDER = 'ASF'

AUTH_DOMAINS = ['asf.alaska.edu', 'earthdata.nasa.gov'] #, 'earthdatacloud.nasa.gov']
AUTH_COOKIES = ['urs_user_already_logged', 'uat_urs_user_already_logged', 'asf-urs']

ERROR_REPORTING_ENDPOINT = 'search-error-report.asf.alaska.edu'



from shapely import unary_union, multipolygons
from typing import Dict, Tuple, Union
from asf_search import ASFSearchOptions, ASFSession, ASFStackableProduct
from asf_search.CMR.translate import try_parse_frame_coverage, try_parse_bool, try_parse_int
from shapely.geometry import shape, MultiPolygon
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform
class NISARProduct(ASFStackableProduct):
    """
    Used for NISAR dataset products

    ASF Dataset Documentation Page: https://asf.alaska.edu/nisar/
    """
    _base_properties = {
        **ASFStackableProduct._base_properties,
        'frameNumber': {
            'path': ['AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0],
            'cast': try_parse_int,
        },  # Sentinel, ALOSm and NISAR product alt for frameNumber (ESA_FRAME)
        'pathNumber': { # AKA trackNumber, AKA relativeOrbit
            'path': ['AdditionalAttributes', ('Name', 'TRACK_NUMBER'), 'Values', 0],
            'cast': try_parse_int,
        },
        'pgeVersion': {'path': ['PGEVersionClass', 'PGEVersion']},
        'crid': {'path': ['DataGranule', 'Identifiers', ('IdentifierType', 'CRID'), 'Identifier']},
        'mainBandPolarization': {'path': ['AdditionalAttributes', ('Name', 'FREQUENCY_A_POLARIZATION'), 'Values']},
        'sideBandPolarization': {'path': ['AdditionalAttributes', ('Name', 'FREQUENCY_B_POLARIZATION'), 'Values']},
        'frameCoverage': {'path': ['AdditionalAttributes', ('Name', 'FULL_FRAME'), 'Values', 0], 'cast': try_parse_frame_coverage},
        'jointObservation': {'path': ['AdditionalAttributes', ('Name', 'JOINT_OBSERVATION'), 'Values', 0], 'cast': try_parse_bool},
        'rangeBandwidth': {'path': ['AdditionalAttributes', ('Name', 'RANGE_BANDWIDTH_CONCAT'), 'Values']},
        'productionConfiguration': {'path': ['AdditionalAttributes', ('Name', 'PRODUCTION_PIPELINE'), 'Values', 0]},
        'processingLevel': {'path': ['AdditionalAttributes', ('Name', 'PRODUCT_TYPE'), 'Values', 0]},
        'bytes': {'path': ['DataGranule', 'ArchiveAndDistributionInformation']},
        'collectionName': {'path': ["CollectionReference", "ShortName"]},
        'orbitType': {'path': ['AdditionalAttributes', ('Name', 'ORBIT_TYPE'), 'Values', 0]},
    }
    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)
        if self.properties.get('processingLevel') is None:
            self.properties.pop('processingLevel', None)

        self.properties['additionalUrls'] = self._get_additional_urls()
        self.properties['browse'] = [url for url in self._get_urls() if url.endswith('.png') or url.endswith('.jpg') or url.endswith('.jpeg')]
        self.properties['s3Urls'] = self._get_s3_uris()

        if self.properties.get('groupID') is None:
            self.properties['groupID'] = self.properties['sceneName']
        self.properties['bytes'] = {
            entry['Name']: {'bytes': entry['SizeInBytes'], 'format': entry['Format']}
            for entry in self.properties['bytes']
        }
        self.properties["conceptID"] = self.umm_get(self.meta, "collection-concept-id")

    @staticmethod
    def get_default_baseline_product_type() -> Union[str, None]:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return None

    def is_valid_reference(self):
        return False

    def get_stack_opts(self, opts: ASFSearchOptions = None) -> ASFSearchOptions:
        """
        Build search options that can be used to find an insar stack for this product

        :return: ASFSearchOptions describing appropriate options
        for building a stack from this product
        """
        return None

    def get_sort_keys(self) -> Tuple[str, str]:
        keys = super().get_sort_keys()

        if keys[0] == '':
            return (self._read_property('processingDate', ''), keys[1])

        return keys


    def _get_geometry(self, item: Dict) -> dict:
        """Overload for dateline multipolygon parsing.
        # TODO consider implications of moving this to base ASFProduct class in future
        """
        try:
            polygons = item['umm']['SpatialExtent']['HorizontalSpatialDomain']['Geometry'][
                'GPolygons'
            ]
            # dateline spanning scenes are stored as multiple polygons in CMR, 
            # we need to unwrap and merge them
            if len(polygons) > 1:
                polygon_shapes = []
                for polygon in polygons:
                    coordinates = [[c['Longitude'], c['Latitude']] for c in polygon['Boundary']['Points']]
                    geometry = self._get_unwrapped({'coordinates': [coordinates], 'type': 'Polygon'})

                    polygon_shapes.append(geometry)
                
                geom = unary_union(multipolygons(polygon_shapes))

                # sometimes the dateline spanning polygons don't overlap properly
                if isinstance(geom, MultiPolygon):
                    geom = geom.convex_hull

                return {'coordinates': [geom.exterior.coords], 'type': 'Polygon'}
            else:
                coordinates = polygons[0]['Boundary']['Points']
                coordinates = [[c['Longitude'], c['Latitude']] for c in coordinates]
                geometry = {'coordinates': [coordinates], 'type': 'Polygon'}
        except KeyError:
            geometry = {'coordinates': None, 'type': 'Polygon'}

        return geometry

    def _get_unwrapped(self, geometry: dict) -> BaseGeometry:
        def unwrap_shape(x, y, z=None):
            x = x if x > 0 else x + 360
            return tuple([x, y])
        wrapped = shape(geometry)
        if wrapped.bounds[0] < 0 or wrapped.bounds[2] < 0:
            unwrapped = transform(unwrap_shape, wrapped)
        else:
            unwrapped = wrapped

        return unwrapped



from typing import Sequence
from copy import copy

from asf_search.search import search
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults


def product_search(product_list: Sequence[str], opts: ASFSearchOptions = None) -> ASFSearchResults:
    """
    Performs a product ID search using the ASF SearchAPI

    Parameters
    ----------
    :param product_list:
        List of specific products.
        Guaranteed to be at most one product per product name.
    opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.

    Returns
    -------
    `asf_search.ASFSearchResults` (list of search results of subclass ASFProduct)
    """

    opts = ASFSearchOptions() if opts is None else copy(opts)

    opts.merge_args(product_list=product_list)

    return search(opts=opts)



SENTINEL1 = 'SENTINEL-1'
SENTINEL1A = 'Sentinel-1A'
SENTINEL1B = 'Sentinel-1B'
SENTINEL1C = 'Sentinel-1C'
SENTINEL1D = 'Sentinel-1D'
SIRC = 'SIR-C'
ALOS = 'ALOS'
ERS = 'ERS'
ERS1 = 'ERS-1'
ERS2 = 'ERS-2'
JERS = 'JERS-1'
RADARSAT = 'RADARSAT-1'
AIRSAR = 'AIRSAR'
SEASAT = 'SEASAT 1'
SMAP = 'SMAP'
UAVSAR = 'UAVSAR'
NISAR = 'NISAR'



from asf_search.ASFProduct import ASFProduct
from asf_search import ASFSession
from asf_search.CMR.translate import try_parse_date
from asf_search.constants import PRODUCT_TYPE


class TROPOProduct(ASFProduct):
    """
    S1Product Subclass made specifically for Sentinel-1 SLC-BURST products

    Key features/properties:
    - `properties['burst']` contains SLC-BURST Specific fields
        such as `fullBurstID` and `burstIndex`
    - `properties['additionalUrls']` contains BURST-XML url
    - SLC-BURST specific stacking params

    ASF Dataset Documentation Page:
        https://asf.alaska.edu/datasets/data-sets/derived-data-sets/sentinel-1-bursts/
    """

    _base_properties = {
        **ASFProduct._base_properties,
        'processingLevel': {
            'path': ['AdditionalAttributes', ('Name', 'PRODUCT_TYPE'), 'Values', 0]
        },
        'bytes': {'path': ['DataGranule', 'ArchiveAndDistributionInformation']},
        'collectionName': {'path': ["CollectionReference", "ShortName"]},
    }

    def __init__(self, args: dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

        if self.properties['processingLevel'] == PRODUCT_TYPE.TROPO_ZENITH:
            self.umm_get(self.umm, 'AdditionalAttributes', ('Name', 'PRODUCT_VERSION'), 'values')
        elif self.properties['processingLevel'] == PRODUCT_TYPE.ECMWF_TROPO:
            self.properties['startTime'] = self.umm_cast(try_parse_date, self.umm_get(self.umm, 'TemporalExtent', 'SingleDateTime'))
            self.properties['stopTime'] = self.properties['startTime']

        bytes_mapping = {
            entry['Name']: {'bytes': entry['SizeInBytes'], 'format': entry['Format']}
            for entry in self.properties['bytes']
        }
        md5sum_mapping = {
            entry['Name']: entry['Checksum']['Value'] for entry in self.properties['bytes']
        }

        self.properties['bytes'] = bytes_mapping
        self.properties['md5sum'] = md5sum_mapping

        self.properties['browse'] = [url for url in self._get_urls() if url.endswith('.png') or url.endswith('.jpg') or url.endswith('.jpeg')]
        self.properties['additionalUrls'] = self._get_additional_urls()
        self.properties['s3Urls'] = self._get_s3_uris()
        
        self.properties["conceptID"] = self.umm_get(self.meta, "collection-concept-id")



import copy
from typing import Dict, List, Optional, Tuple
from asf_search import ASFSearchOptions, ASFSession, ASFStackableProduct
from asf_search.CMR.translate import try_parse_date
from asf_search.CMR.translate import try_parse_int
from asf_search.constants import PLATFORM
from asf_search.constants import PRODUCT_TYPE


class S1Product(ASFStackableProduct):
    """
    The S1Product classes covers most Sentinel-1 Products
    (For S1 BURST-SLC, OPERA-S1, and ARIA-S1 GUNW Products, see relevant S1 subclasses)

    ASF Dataset Overview Page: https://asf.alaska.edu/datasets/daac/sentinel-1/
    """

    _base_properties = {
        **ASFStackableProduct._base_properties,
        'frameNumber': {
            'path': ['AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0],
            'cast': try_parse_int,
        },  # Sentinel and ALOS product alt for frameNumber (ESA_FRAME)
        'groupID': {'path': ['AdditionalAttributes', ('Name', 'GROUP_ID'), 'Values', 0]},
        'md5sum': {'path': ['AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0]},
        'pgeVersion': {'path': ['PGEVersionClass', 'PGEVersion']},
    }
    """
    S1 Specific path override
    - frameNumber: overrides ASFProduct's `CENTER_ESA_FRAME` with `FRAME_NUMBER`
    """

    baseline_type = ASFStackableProduct.BaselineCalcType.CALCULATED

    def __init__(self, args: Dict = {}, session: ASFSession = ASFSession()):
        super().__init__(args, session)

        self.properties['s3Urls'] = self._get_s3_uris()

        if self.has_baseline():
            self.baseline = self.get_baseline_calc_properties()

    def has_baseline(self) -> bool:
        baseline = self.get_baseline_calc_properties()

        return baseline is not None and None not in baseline['stateVectors']['positions'].values()

    def get_baseline_calc_properties(self) -> Dict:
        """
        :returns properties required for SLC baseline stack calculations
        """
        ascendingNodeTime = self.umm_cast(
            self._parse_timestamp,
            self.umm_get(self.umm, 'AdditionalAttributes', ('Name', 'ASC_NODE_TIME'), 'Values', 0),
        )

        return {
            'stateVectors': self.get_state_vectors(),
            'ascendingNodeTime': ascendingNodeTime,
        }

    def get_state_vectors(self) -> Dict:
        """
        Used in spatio-temporal perpendicular baseline calculations for non-pre-calculated stacks

        :returns dictionary of pre/post positions, velocities, and times"""
        positions = {}
        velocities = {}

        sv_pre_position = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'SV_POSITION_PRE'), 'Values', 0
        )
        sv_post_position = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'SV_POSITION_POST'), 'Values', 0
        )
        sv_pre_velocity = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'SV_VELOCITY_PRE'), 'Values', 0
        )
        sv_post_velocity = self.umm_get(
            self.umm, 'AdditionalAttributes', ('Name', 'SV_VELOCITY_POST'), 'Values', 0
        )

        positions['prePosition'], positions['prePositionTime'] = self.umm_cast(
            self._parse_state_vector, sv_pre_position
        )
        positions['postPosition'], positions['postPositionTime'] = self.umm_cast(
            self._parse_state_vector, sv_post_position
        )
        velocities['preVelocity'], velocities['preVelocityTime'] = self.umm_cast(
            self._parse_state_vector, sv_pre_velocity
        )
        velocities['postVelocity'], velocities['postVelocityTime'] = self.umm_cast(
            self._parse_state_vector, sv_post_velocity
        )

        return {'positions': positions, 'velocities': velocities}

    def _parse_timestamp(self, timestamp: str) -> Optional[str]:
        if timestamp is None:
            return None

        return try_parse_date(timestamp)

    def _parse_state_vector(self, state_vector: str) -> Tuple[Optional[List], Optional[str]]:
        if state_vector is None:
            return None, None

        velocity = [float(val) for val in state_vector.split(',')[:3]]
        timestamp = self._parse_timestamp(state_vector.split(',')[-1])

        return velocity, timestamp

    def get_stack_opts(self, opts: ASFSearchOptions = None) -> ASFSearchOptions:
        """
        Returns the search options asf-search will use internally
        to build an SLC baseline stack from

        :param opts: additional criteria for limiting
        :returns ASFSearchOptions used for build Sentinel-1 SLC Stack
        """
        stack_opts = ASFSearchOptions() if opts is None else copy(opts)

        stack_opts.processingLevel = self.get_default_baseline_product_type()
        stack_opts.beamMode = [self.properties['beamModeType']]
        stack_opts.flightDirection = self.properties['flightDirection']
        stack_opts.relativeOrbit = [int(self.properties['pathNumber'])]  # path
        stack_opts.platform = [PLATFORM.SENTINEL1A, PLATFORM.SENTINEL1B, PLATFORM.SENTINEL1C, PLATFORM.SENTINEL1D,]

        if self.properties['polarization'] in ['HH', 'HH+HV']:
            stack_opts.polarization = ['HH', 'HH+HV']
        else:
            stack_opts.polarization = ['VV', 'VV+VH']

        stack_opts.intersectsWith = self.centroid().wkt

        return stack_opts

    def is_valid_reference(self) -> bool:
        keys = ['postPosition', 'postPositionTime', 'prePosition', 'postPositionTime']

        for key in keys:
            if self.baseline['stateVectors']['positions'].get(key) is None:
                return False

        return True

    @staticmethod
    def get_default_baseline_product_type() -> str:
        """
        Returns the product type to search for when building a baseline stack.
        """
        return PRODUCT_TYPE.SLC



import time
from typing import Dict, Generator, Literal, Union, Sequence, Tuple, List
from copy import copy
from requests.exceptions import HTTPError
from requests import ReadTimeout, Response
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
    wait_fixed,
)
import datetime
import dateparser

from asf_search import ASF_LOGGER

from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR import build_subqueries, translate_opts
from asf_search.CMR.datasets import dataset_collections

from asf_search.ASFSession import ASFSession
from asf_search.ASFProduct import ASFProduct
from asf_search.exceptions import (
    ASFSearch4xxError,
    ASFSearch5xxError,
    ASFSearchError,
    CMRIncompleteError,
)
from asf_search.constants import INTERNAL
from asf_search.WKT.validate_wkt import validate_wkt
from asf_search.search.error_reporting import report_search_error
import asf_search.Products as ASFProductType


def search_generator(
    absoluteOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    asfFrame: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    beamMode: Union[str, Sequence[str]] = None,
    beamSwath: Union[str, Sequence[str]] = None,
    campaign: Union[str, Sequence[str]] = None,
    maxDoppler: float = None,
    minDoppler: float = None,
    end: Union[datetime.datetime, str] = None,
    maxFaradayRotation: float = None,
    minFaradayRotation: float = None,
    flightDirection: str = None,
    flightLine: str = None,
    frame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
    granule_list: Union[str, Sequence[str]] = None,
    groupID: Union[str, Sequence[str]] = None,
    insarStackId: str = None,
    instrument: Union[str, Sequence[str]] = None,
    intersectsWith: str = None,
    lookDirection: Union[str, Sequence[str]] = None,
    offNadirAngle: Union[
        float, Tuple[float, float], Sequence[Union[float, Tuple[float, float]]]
    ] = None,
    platform: Union[str, Sequence[str]] = None,
    polarization: Union[str, Sequence[str]] = None,
    processingDate: Union[datetime.datetime, str] = None,
    processingLevel: Union[str, Sequence[str]] = None,
    product_list: Union[str, Sequence[str]] = None,
    relativeOrbit: Union[
        int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]
    ] = None,
    season: Tuple[int, int] = None,
    start: Union[datetime.datetime, str] = None,
    ariaVersion: str = None,
    absoluteBurstID: Union[int, Sequence[int]] = None,
    relativeBurstID: Union[int, Sequence[int]] = None,
    fullBurstID: Union[str, Sequence[str]] = None,
    temporalBaselineDays: Union[str, Sequence[str]] = None,
    operaBurstID: Union[str, Sequence[str]] = None,
    frameCoverage: Literal['FULL', 'PARTIAL'] = None,
    mainBandPolarization: Union[str, Sequence[str]] = None,
    sideBandPolarization: Union[str, Sequence[str]] = None,
    rangeBandwidth: Union[str, Sequence[str]] = None,
    jointObservation: bool = None,
    productionConfiguration: Union[Literal['PR', 'UR'], Sequence[Literal['PR', 'UR']]] = None,
    dataset: Union[str, Sequence[str]] = None,
    collections: Union[str, Sequence[str]] = None,
    shortName: Union[str, Sequence[str]] = None,
    cmr_keywords: Union[Tuple[str, str], Sequence[Tuple[str, str]]] = None,
    tileID: Union[str, Sequence[str]] = None,
    maxResults: int = None,
    opts: ASFSearchOptions = None,
) -> Generator[ASFSearchResults, None, None]:
    """
    Performs a generic search against the Central Metadata Repository (CMR),
    yielding results page by page (250 products at a time by default) as they're returned from CMR.

    Accepts a number of search parameters, and/or an ASFSearchOptions object.
    If an ASFSearchOptions object is provided as well as other specific parameters,
    the two sets of options will be merged, preferring the specific keyword arguments.

    Parameters
    ----------
    absoluteOrbit:
        For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B
        this value corresponds to the orbit count within the orbit cycle.
        For UAVSAR it is the Flight ID.
    ariaVersion:
        For ARIAS1GUNW this value describes the version a product was generated under
    asfFrame:
        This is primarily an ASF / JAXA frame reference. However,
        some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    beamMode:
        The beam mode used to acquire the data.
    beamSwath:
        Encompasses a look angle and beam mode.
    campaign:
        For UAVSAR and AIRSAR data collections only. Search by general location,
        site description, or data grouping as supplied by flight agency or project.
    maxDoppler:
        Doppler provides an indication of how much the look direction deviates
        from the ideal perpendicular flight direction acquisition.
    minDoppler:
        Doppler provides an indication of how much the look direction deviates
        from the ideal perpendicular flight direction acquisition.
    end:
        End date of data acquisition. Supports timestamps
        as well as natural language such as "3 weeks ago"
    maxFaradayRotation:
        Rotation of the polarization plane of
        the radar signal impacts imagery, as HH and HV signals become mixed.
    minFaradayRotation:
        Rotation of the polarization plane of
        the radar signal impacts imagery, as HH and HV signals become mixed.
    flightDirection:
        Satellite orbit direction during data acquisition
    flightLine:
        Specify a flightline for UAVSAR or AIRSAR.
    frame:
        ESA-referenced frames are offered to give users a universal framing convention.
        Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    granule_list:
        List of specific granules.
        Search results may include several products per granule name.
        Supports wildcard queries (*/?)
    groupID:
        Identifier used to find products considered to
        be of the same scene but having different granule names
    insarStackId:
        Identifier used to find products of the same InSAR stack
    instrument:
        The instrument used to acquire the data. See also: platform
    intersectsWith:
        Search by polygon, linestring,
        or point defined in 2D Well-Known Text (WKT)
    lookDirection:
        Left or right look direction during data acquisition
    offNadirAngle:
        Off-nadir angles for ALOS PALSAR
    platform:
        Remote sensing platform that acquired the data.
        Platforms that work together, such as Sentinel-1A/1B and ERS-1/2
        have multi-platform aliases available. See also: instrument
    polarization:
        A property of SAR electromagnetic waves
        that can be used to extract meaningful information about surface properties of the earth.
    processingDate:
        Used to find data that has been processed at ASF since a given
        time and date. Supports timestamps as well as natural language such as "3 weeks ago"
    processingLevel:
        Level to which the data has been processed
    product_list:
        List of specific products.
        Guaranteed to be at most one product per product name.
    relativeOrbit:
        Path or track of satellite during data acquisition.
        For UAVSAR it is the Line ID.
    season:
        Start and end day of year for desired seasonal range.
        This option is used in conjunction with start/end to specify a seasonal range
        within an overall date range.
    start:
        Start date of data acquisition.
        Supports timestamps as well as natural language such as "3 weeks ago"
    tileID:
        For DIST-ALERT-S1 product type products
    collections:
        List of collections (concept-ids) to limit search to
    temporalBaselineDays:
        List of temporal baselines,
        used for Sentinel-1 Interferogram (BETA)
    maxResults:
        The maximum number of results to be returned by the search
    opts:
        An ASFSearchOptions object describing the search parameters to be used.
        Search parameters specified outside this object will override in event of a conflict.

    Yields
    -------
    `asf_search.ASFSearchResults` (list of search results of subclass ASFProduct, page by page)
    """

    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = ASFSearchOptions() if kwargs['opts'] is None else copy(opts)
    del kwargs['opts']

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    maxResults = opts.pop('maxResults', None)

    if maxResults is not None and (
        getattr(opts, 'granule_list', False) or getattr(opts, 'product_list', False)
    ):
        names = [] if opts.granule_list is None else opts.granule_list
        if not any('*' in name or '?' in name for name in names):
            raise ValueError(
                'Cannot use maxResults with product_list or non-wildcard granule_list.'
            )

    ASF_LOGGER.debug(f'SEARCH: preprocessing opts: {opts}')
    preprocess_opts(opts)
    ASF_LOGGER.debug(f'SEARCH: preprocessed opts: {opts}')

    ASF_LOGGER.info(f'SEARCH: Using search opts {opts}')

    url = '/'.join(s.strip('/') for s in [f'https://{opts.host}', f'{INTERNAL.CMR_GRANULE_PATH}'])
    total = 0

    queries = build_subqueries(opts)

    ASF_LOGGER.info(f'SEARCH: Using cmr endpoint: "{url}"')
    ASF_LOGGER.debug(f'SEARCH: Built {len(queries)} subqueries')

    for subquery_idx, query in enumerate(queries):
        ASF_LOGGER.info(f'SUBQUERY {subquery_idx + 1}: Beginning subquery with opts: {query}')

        ASF_LOGGER.debug(f'TRANSLATION: Translating subquery:\n{query}')
        translated_opts = translate_opts(query)
        ASF_LOGGER.debug(f'TRANSLATION: Subquery translated to cmr keywords:\n{translated_opts}')
        cmr_search_after_header = ''
        subquery_count = 0

        page_number = 1
        while cmr_search_after_header is not None:
            try:
                ASF_LOGGER.debug(f'SUBQUERY {subquery_idx + 1}: Fetching page {page_number}')
                items, subquery_max_results, cmr_search_after_header = query_cmr(
                    opts.session, url, translated_opts, subquery_count
                )
            except ASFSearchError as exc:
                message = str(exc)
                ASF_LOGGER.error(message)
                report_search_error(query, message)
                opts.session.headers.pop('CMR-Search-After', None)
                raise

            ASF_LOGGER.debug(
                f'SUBQUERY {subquery_idx + 1}: Page {page_number} fetched, returned {len(items)} items.'
            )

            if (
                len(items) != INTERNAL.CMR_PAGE_SIZE
                and len(items) + subquery_count < subquery_max_results
            ):
                message = (
                    'CMR returned page of incomplete results.'
                    f'Expected {min(INTERNAL.CMR_PAGE_SIZE, subquery_max_results - subquery_count)} results,'
                    f'got {len(items)}'
                )
                ASF_LOGGER.warning(message)
                report_search_error(query, message)

            opts.session.headers.update({'CMR-Search-After': cmr_search_after_header})
            perf = time.time()
            last_page = process_page(
                items, maxResults, subquery_max_results, total, subquery_count, opts
            )
            ASF_LOGGER.info(f'Page Processing Time {time.time() - perf}')
            subquery_count += len(last_page)
            total += len(last_page)
            last_page.searchComplete = subquery_count == subquery_max_results or total == maxResults
            yield last_page

            if last_page.searchComplete:
                if total == maxResults:  # the user has as many results as they wanted
                    ASF_LOGGER.info(f'SEARCH COMPLETE: MaxResults ({maxResults}) reached')
                    opts.session.headers.pop('CMR-Search-After', None)
                    return
                else:  # or we've gotten all possible results for this subquery
                    ASF_LOGGER.info(
                        f'SUBQUERY {subquery_idx + 1} COMPLETE: results exhausted for subquery'
                    )
                    cmr_search_after_header = None

            page_number += 1

        opts.session.headers.pop('CMR-Search-After', None)

    ASF_LOGGER.info(f'SEARCH COMPLETE: results exhausted for search opts {opts}')


@retry(
    reraise=True,
    retry=retry_if_exception_type(CMRIncompleteError),
    wait=wait_fixed(2),
    stop=stop_after_attempt(3),
)
def query_cmr(
    session: ASFSession,
    url: str,
    translated_opts: Dict,
    sub_query_count: int,
):
    response = get_page(session=session, url=url, translated_opts=translated_opts)

    perf = time.time()
    items = [as_ASFProduct(f, session=session) for f in response.json()['items']]
    ASF_LOGGER.debug(f'Product Subclassing Time {time.time() - perf}')
    hits: int = response.json()['hits']  # total count of products given search opts
    # 9-10 per process
    # 3.9-5 per process
    # sometimes CMR returns results with the wrong page size
    # if len(items) != INTERNAL.CMR_PAGE_SIZE and len(items) + sub_query_count < hits:
    #     raise CMRIncompleteError(
    #         'CMR returned page of incomplete results.'
    #         f'Expected {min(INTERNAL.CMR_PAGE_SIZE, hits - sub_query_count)} results,'
    #         f'got {len(items)}'
    #     )

    return items, hits, response.headers.get('CMR-Search-After', None)


def process_page(
    items: List[ASFProduct],
    max_results: int,
    subquery_max_results: int,
    total: int,
    subquery_count: int,
    opts: ASFSearchOptions,
):
    if max_results is None:
        last_page = ASFSearchResults(
            items[: min(subquery_max_results - subquery_count, len(items))], opts=opts
        )
    else:
        last_page = ASFSearchResults(items[: min(max_results - total, len(items))], opts=opts)
    return last_page


@retry(
    reraise=True,
    retry=retry_if_exception_type(ASFSearch5xxError),
    wait=wait_exponential(
        multiplier=1, min=3, max=10
    ),  # Wait 2^x * 1 starting with 3 seconds, max 10 seconds between retries
    stop=stop_after_attempt(3),
)
def get_page(session: ASFSession, url: str, translated_opts: List) -> Response:
    from asf_search.constants.INTERNAL import CMR_TIMEOUT

    perf = time.time()
    try:
        response = session.post(url=url, data=translated_opts, timeout=CMR_TIMEOUT)
        response.raise_for_status()
    except HTTPError as exc:
        error_message = f'HTTP {response.status_code}: {response.json()["errors"]}'
        if 400 <= response.status_code <= 499:
            raise ASFSearch4xxError(error_message) from exc
        if 500 <= response.status_code <= 599:
            raise ASFSearch5xxError(error_message) from exc
    except ReadTimeout as exc:
        raise ASFSearchError(
            f'Connection Error (Timeout): CMR took too long to respond. Set asf constant "asf_search.constants.INTERNAL.CMR_TIMEOUT" to increase. ({url=}, timeout={CMR_TIMEOUT})'
        ) from exc

    ASF_LOGGER.info(f'Query Time Elapsed {time.time() - perf}')
    return response


def preprocess_opts(opts: ASFSearchOptions):
    # Repair WKT here so it only happens once, and you can save the result to the new Opts object:
    wrap_wkt(opts=opts)

    # Date/Time logic, convert "today" to the literal timestamp if needed:
    set_default_dates(opts=opts)

    # Platform Alias logic:
    set_platform_alias(opts=opts)

    set_science_product_alias(opts=opts)


def wrap_wkt(opts: ASFSearchOptions):
    if opts.intersectsWith is not None:
        wrapped, _, repairs = validate_wkt(opts.intersectsWith)
        opts.intersectsWith = wrapped.wkt
        if len(repairs):
            ASF_LOGGER.warning(
                'WKT REPAIR/VALIDATION: The following repairs were performed '
                f'on the provided AOI:\n{[str(repair) for repair in repairs]}'
            )


def set_default_dates(opts: ASFSearchOptions):
    if opts.start is not None and isinstance(opts.start, str):
        opts.start = dateparser.parse(opts.start, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    if opts.end is not None and isinstance(opts.end, str):
        opts.end = dateparser.parse(opts.end, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    # If both are used, make sure they're in the right order:
    if opts.start is not None and opts.end is not None:
        if opts.start > opts.end:
            ASF_LOGGER.warning(
                f'Start date ({opts.start}) is after end date ({opts.end}). Switching the two.'
            )
            opts.start, opts.end = opts.end, opts.start
    # Can't do this sooner, since you need to compare start vs end:
    # TODO python 3.13 deprecation warning, ambiguous date when month specified but not year
    if opts.start is not None:
        opts.start = opts.start.strftime('%Y-%m-%dT%H:%M:%SZ')
    if opts.end is not None:
        opts.end = opts.end.strftime('%Y-%m-%dT%H:%M:%SZ')


def set_platform_alias(opts: ASFSearchOptions):
    # Platform Alias logic:
    if opts.platform is not None:
        plat_aliases = {
            # Groups:
            'S1': ['SENTINEL-1A', 'SENTINEL-1B', 'SENTINEL-1C', 'SENTINEL-1D'],
            'SENTINEL-1': ['SENTINEL-1A', 'SENTINEL-1B', 'SENTINEL-1C', 'SENTINEL-1D'],
            'SENTINEL': ['SENTINEL-1A', 'SENTINEL-1B', 'SENTINEL-1C', 'SENTINEL-1D'],
            'ERS': ['ERS-1', 'ERS-2'],
            'SIR-C': ['STS-59', 'STS-68'],
            # Singles / Aliases:
            'R1': ['RADARSAT-1'],
            'E1': ['ERS-1'],
            'E2': ['ERS-2'],
            'J1': ['JERS-1'],
            'A3': ['ALOS'],
            'AS': ['DC-8'],
            'AIRSAR': ['DC-8'],
            'SS': ['SEASAT 1'],
            'SEASAT': ['SEASAT 1'],
            'SA': ['SENTINEL-1A'],
            'SB': ['SENTINEL-1B'],
            'SC': ['SENTINEL-1C'],
            'SD': ['SENTINEL-1D'],
            'SP': ['SMAP'],
            'UA': ['G-III'],
            'UAVSAR': ['G-III'],
        }
        platform_list = []
        for plat in opts.platform:
            # If it's a key, *replace* it with all the values. Else just add the key:
            if plat.upper() in plat_aliases:
                platform_list.extend(plat_aliases[plat.upper()])
            else:
                platform_list.append(plat)

        opts.platform = list(set(platform_list))


_dataset_collection_items = dataset_collections.items()


def set_science_product_alias(opts: ASFSearchOptions):
    """Alias certain product types (primarily NISAR L0B)"""
    if opts.processingLevel is not None:
        processingLevelAliases = {'L0B': ['RRSD']}

        processing_levels = []
        for processing_level in opts.processingLevel:
            # If it's a key, *replace* it with all the values. Else just add the key:
            if processing_level.upper() in processingLevelAliases:
                processing_levels.extend(processingLevelAliases[processing_level.upper()])
            else:
                processing_levels.append(processing_level)

        opts.processingLevel = processing_levels


def as_ASFProduct(item: Dict, session: ASFSession) -> ASFProduct:
    """Returns the granule umm as the corresponding ASFProduct subclass,
    or ASFProduct if no equivalent is found

    :param item: the granule umm json
    :param session: the session used to query CMR for the product

    :returns the granule as an object of type ASFProduct
    """
    if ASFProductType.OPERAS1Product._is_subclass(item):
        return ASFProductType.OPERAS1Product(item, session=session)

    product_type_key = _get_product_type_key(item)

    # if there's a direct entry in our dataset to product type dict
    # perf = time.time()
    subclass = dataset_to_product_types.get(product_type_key)
    if subclass is not None:
        # ASF_LOGGER.warning(f'subclass selection time {time.time() - perf}')
        return subclass(item, session=session)

    # if the key matches one of the shortnames in any of our datasets
    for dataset, collections in _dataset_collection_items:
        if collections.get(product_type_key) is not None:
            subclass = dataset_to_product_types.get(dataset)
            if subclass is not None:
                # ASF_LOGGER.warning(f'subclass selection time {time.time() - perf}')
                return subclass(item, session=session)
            break  # dataset exists, but is not in dataset_to_product_types yet

    # If the platform exists, try to match it
    platform = _get_platform(item=item)
    if ASFProductType.ARIAS1GUNWProduct._is_subclass(item=item):
        return dataset_to_product_types.get('ARIA S1 GUNW')(item, session=session)
    elif (subclass := dataset_to_product_types.get(platform)) is not None:
        return subclass(item, session=session)

    output = ASFProduct(item, session=session)

    granule_concept_id = output.meta.get('concept-id', 'Missing Granule Concept ID')
    fileID = output.properties.get(
        'fileID', output.properties.get('sceneName', 'fileID and sceneName Missing')
    )

    ASF_LOGGER.warning(
        f'Failed to find corresponding ASFProduct subclass for \
        Product: "{fileID}", Granule Concept ID: "{granule_concept_id}", \
        default to "ASFProduct"'
    )
    return output


def _get_product_type_key(item: Dict) -> str:
    """Match the umm response to the right ASFProduct subclass by returning one of the following:
    1. collection shortName (Ideal case)
    2. platform_shortName (Fallback)
        - special case: Aria S1 GUNW
    """
    collection_shortName = ASFProduct.umm_get(item['umm'], 'CollectionReference', 'ShortName')

    if collection_shortName is None:
        if ASFProductType.ARIAS1GUNWProduct._is_subclass(item=item):
            return 'ARIA S1 GUNW'

        platform = _get_platform(item=item)
        return platform

    return collection_shortName


def _get_platform(item: Dict):
    return ASFProduct.umm_get(item['umm'], 'Platforms', 0, 'ShortName')


# Maps datasets from DATASET.py and collection/platform shortnames to ASFProduct subclasses
dataset_to_product_types = {
    'SENTINEL-1': ASFProductType.S1Product,
    'OPERA-S1': ASFProductType.OPERAS1Product,
    'OPERA-S1-CALVAL': ASFProductType.OPERAS1Product,
    'SLC-BURST': ASFProductType.S1BurstProduct,
    'ALOS': ASFProductType.ALOSProduct,
    'SIR-C': ASFProductType.SIRCProduct,
    'STS-59': ASFProductType.SIRCProduct,
    'STS-68': ASFProductType.SIRCProduct,
    'ARIA S1 GUNW': ASFProductType.ARIAS1GUNWProduct,
    'SMAP': ASFProductType.SMAPProduct,
    'UAVSAR': ASFProductType.UAVSARProduct,
    'G-III': ASFProductType.UAVSARProduct,
    'RADARSAT-1': ASFProductType.RADARSATProduct,
    'ERS': ASFProductType.ERSProduct,
    'ERS-1': ASFProductType.ERSProduct,
    'ERS-2': ASFProductType.ERSProduct,
    'JERS-1': ASFProductType.JERSProduct,
    'AIRSAR': ASFProductType.AIRSARProduct,
    'DC-8': ASFProductType.AIRSARProduct,
    'SEASAT': ASFProductType.SEASATProduct,
    'SEASAT 1': ASFProductType.SEASATProduct,
    'NISAR': ASFProductType.NISARProduct,
    'ALOS-2': ASFProductType.ALOS2Product,
    'TROPO': ASFProductType.TROPOProduct,
}



# Sentinel-1
GRD_HD = 'GRD_HD'
GRD_MD = 'GRD_MD'
GRD_MS = 'GRD_MS'
GRD_HS = 'GRD_HS'
GRD_FD = 'GRD_FD'
SLC = 'SLC'
OCN = 'OCN'
RAW = 'RAW'
METADATA_GRD_HD = 'METADATA_GRD_HD'
METADATA_GRD_MD = 'METADATA_GRD_MD'
METADATA_GRD_MS = 'METADATA_GRD_MS'
METADATA_GRD_HS = 'METADATA_GRD_HS'
METADATA_SLC = 'METADATA_SLC'
METADATA_OCN = 'METADATA_OCN'
METADATA_RAW = 'METADATA_RAW'
BURST = 'BURST'

# ALOS PALSAR
L1_0 = 'L1.0'
L1_1 = 'L1.1'
L1_5 = 'L1.5'
L2_2 = 'L2.2'
RTC_LOW_RES = 'RTC_LOW_RES'
RTC_HIGH_RES = 'RTC_HI_RES'
KMZ = 'KMZ'

# ALOS AVNIR
# No PROCESSING_TYPE attribute in CMR

# SIR-C
# SLC and SLC metadata are both 'SLC', provided by Sentinel-1 constants

# Sentinel-1 InSAR
GUNW_STD = 'GUNW_STD'
GUNW_AMP = 'GUNW_AMP'
GUNW_CON = 'GUNW_CON'
GUN_COH = 'GUNW_COH'
GUNW_UNW = 'GUNW_UNW'

# SMAP
L1A_RADAR_RO_HDF5 = 'L1A_Radar_RO_HDF5'
L1A_RADAR_HDF5 = 'L1A_Radar_HDF5'
L1B_S0_LOW_RES_HDF5 = 'L1B_S0_LoRes_HDF5'
L1C_S0_HIGH_RES_HDF5 = 'L1C_S0_HiRes_HDF5'
L1A_RADAR_RO_QA = 'L1A_Radar_RO_QA'
L1A_RADAR_QA = 'L1A_Radar_QA'
L1B_S0_LOW_RES_QA = 'L1B_S0_LoRes_QA'
L1C_S0_HIGH_RES_QA = 'L1C_S0_HiRes_QA'
L1A_RADAR_RO_ISO_XML = 'L1A_Radar_RO_ISO_XML'
L1B_S0_LOW_RES_ISO_XML = 'L1B_S0_LoRes_ISO_XML'
L1C_S0_HIGH_RES_ISO_XML = 'L1C_S0_HiRes_ISO_XML'

# UAVSAR
AMPLITUDE = 'AMPLITUDE'
STOKES = 'STOKES'
AMPLITUDE_GRD = 'AMPLITUDE_GRD'
PROJECTED = 'PROJECTED'
PROJECTED_ML5X5 = 'PROJECTED_ML5X5'
PROJECTED_ML3X3 = 'PROJECTED_ML3X3'
INTERFEROMETRY_GRD = 'INTERFEROMETRY_GRD'
INTERFEROMETRY = 'INTERFEROMETRY'
COMPLEX = 'COMPLEX'
# KMZ provided by ALOS PALSAR
INC = 'INC'
SLOPE = 'SLOPE'
DEM_TIFF = 'DEM_TIFF'
PAULI = 'PAULI'
METADATA = 'METADATA'

# RADARSAT
L0 = 'L0'
L1 = 'L1'

# ERS
# L0 provided by RADARSAT
# L1 provided by RADARSAT

# JERS
# L0 provided by RADARSAT
# L1 provided by RADARSAT

# AIRSAR
CTIF = 'CTIF'
PTIF = 'PTIF'
LTIF = 'LTIF'
JPG = 'JPG'
LSTOKES = 'LSTOKES'
PSTOKES = 'PSTOKES'
CSTOKES = 'CSTOKES'
DEM = 'DEM'
THREEFP = '3FP'

# OPERA-S1
RTC = 'RTC'
CSLC = 'CSLC'
RTC_STATIC = 'RTC-STATIC'
CSLC_STATIC = 'CSLC-STATIC'
DISP_S1 = 'DISP-S1'
DISP_S1_STATIC = 'DISP-S1-STATIC'
DIST_ALERT_S1 = 'DIST-ALERT-S1'

# TROPO
TROPO_ZENITH = 'TROPO-ZENITH'
ECMWF_TROPO = 'ECMWF_TROPO'

### NISAR

### NISAR Science Products ###
L0B = 'L0B'
"""alias for RRSD Level 0B product types"""

RRSD = 'RRSD'
"""Level 0B Radar Raw Signal Data"""

RSLC = 'RSLC'
"""Level 1 Range-Doppler Single Look Complex"""
RIFG = 'RIFG'
"""Level 1 Range-Doppler Wrapped Interferrogram"""
RUNW = 'RUNW'
"""Level 1 Range-Doppler Unwrapped Interferrogram"""
ROFF = 'ROFF'
"""Level 1 Range-Doppler Pixel Offsets"""
GSLC = 'GSLC'
"""Level 2 Geocoded Single Look Complex"""
GCOV = 'GCOV'
"""Level 2 Geocoded Polarimetric Covariance"""
GUNW = 'GUNW'
""""Level 2 Geocoded Unwrapped Inteferrogram"""
GOFF = 'GOFF'
"""Level 2 Geocoded Pixel Offsets"""
SME2 = 'SME2'
"""Level 3 Soil Moisture EASE-Grid 2.0"""

### NISAR ANCILLARY PRODUCTS ###

CRSD = 'CRSD'
"""NISAR Radar Raw Signal Calibration Data"""

# NISAR Coordinated Observation Plan (NISAR_COP)
DCOP = 'DCOP'
"""NISAR Coordinates Observation Plan"""
# NISAR_OROST
OROST = 'OROST'
"""NISAR Radar Observation Sequence Table"""
# NISAR_STUF
STUF = "STUF"
"""	STUF - 1 week plan of orbit boundaries, convert UTC to Orbit"""

# NISAR Total Electron Content (NISAR_TEC)
TEC = "TEC"
"NISAR Total Electron Content"

# NISAR L-SAR Radar Clock to UTC Conversion File (NISAR_LRCLK_UTC)
LRCLK_UTC = "LRCLK_UTC"
"""L-SAR Radar clock to UTC Spacecraft Clock Spacecraft Event Time (SCLKSCET) conversion file"""


# NISAR Orbit Ephemeris (NISAR_OE)
# "The NASA-ISRO Synthetic Aperture Radar (NISAR) Orbit Ephemeris collection contains 
# the state vector files for the NISAR mission"
FOE = "FOE"
"""Forecast Orbit Ephemeris"""
MOE = "MOE"
"""	Medium Orbit Ephemeris"""
NOE = "NOE"
"""	Near-Realtime Orbit Ephemeris"""
POE = "POE"
"""	Precise Orbit Ephemeris"""

# NISAR Ancillary and Auxiliary Data (NISAR_ANC_AUX)
# "The NASA-ISRO Synthetic Aperture Radar (NISAR) Ancilliary and Auxiallry collection contains products 
# that are supplementary information for the NISAR mission and are created a limited number of times.""
TSR_STATIC = "TSR_STATIC"
"""Time Series Ratio (TSR_STATIC)"""
PMI_STATIC = "PMI_STATIC"
"""Physical Model Inversion (PMI_STATIC)"""
CORNER_REFL = "CORNER_REFL"
"""Corner Reflector locations (CORNER_REFL)"""
LSAR_INT_CAL = "LSAR_INT_CAL"
"""Combined ADT Flight Table (INT_CAL)"""
LSAR_EXT_CAL = "LSAR_EXT_CAL"
"""LSAR Calibration File (EXT_CAL)"""
FT_WAVEFORM = "FT_WAVEFORM"
"""Flight Table Waveform (FT_WAVEFORM)"""
FT_PARAM = "FT_PARAM"
"""Flight Table Parameters (FT_PARAM)"""
ANTPAT = "ANTPAT"
"""Antenna Pattern (ANTPAT)"""
BFPQ = "BFPQ"
"""Block Floating Point Quantization Look-up Table (BFPQ)"""
TFDB = "TFDB"
"""Track Frame Database (TFDB)"""
L_CHAN_DATA = "L_CHAN_DATA"
"""L-SAR Channel Data (L_CHAN_DATA)"""
SM_STATIC = "SM_STATIC"
"""Soil Moisture Static File (SM_STATIC)"""
DSG_STATIC = "DSG_STATIC"
"""Disaggregation File (DSG_STATIC)"""

# NISAR Pointing: (NISAR_RP)
# "The NASA-ISRO Synthetic Aperture Radar (NISAR) Radar Pointing collection contains the radar pointing files for the NISAR mission.
# The collection includes multiple accuracy levels"
FRP = "FRP"
"""	Forecast Radar Pointing"""
PRP = "PRP"
"""	Precise Radar Pointing"""
NRP = "NRP"
"""	Near-Realtime Radar Pointing"""

# ECMWF SMST Data (ASF_ECMWF_SMST)
ECMWF_SMST = "ECMWF_SMST"
"""	European Centre for Medium-Range Weather Forecasts (models)"""

# TODO: Verify/Add/Remove these product types as necessary
# COP = "COP"
# """Coordinated Observation Plan that contains a 7 day plan of observations"""
# DEM = "DEM"
# """	Digital Elevation Model"""

# LOCAL_INC_ANG = "LOCAL_INC_ANG" 
# """	Local Incidence Angles"""
# VWC = "VWC"
# """	Vegetation water content"""
# WATER_MASK = "WaterMask"
# """Water Mask"""



from typing import List
from numbers import Number
from asf_search.baseline.stack import find_new_reference, get_baseline_from_stack
from asf_search import ASFProduct, ASFSearchResults, ASFSession, ASFStackableProduct
from asf_search.search.search_generator import as_ASFProduct

import pytest


def run_test_find_new_reference(stack: List, output_index: Number) -> None:
    """
    Test asf_search.baseline.stack.find_new_reference
    """

    if stack == []:
        assert find_new_reference(stack) is None
    else:
        products = [as_ASFProduct(product, ASFSession()) for product in stack]
        for idx, product in enumerate(products):
            product = clear_baseline(stack[idx], product)
        assert (
            find_new_reference(products).properties['sceneName']
            == stack[output_index]['properties']['sceneName']
        )


def run_test_get_default_product_type(product: ASFStackableProduct, product_type: str) -> None:
    assert product.get_default_baseline_product_type() == product_type


def run_test_get_baseline_from_stack(reference, stack, output_stack, error):
    reference = as_ASFProduct(reference, ASFSession())
    stack = ASFSearchResults([as_ASFProduct(product, ASFSession()) for product in stack])

    if error is None:
        stack, warnings = get_baseline_from_stack(reference, stack)

        keys = ['sceneName', 'perpendicularBaseline', 'temporalBaseline']

        for idx, product in enumerate(stack):
            for key in keys:
                assert product.properties[key] == output_stack[idx]['properties'][key]

        return

    with pytest.raises(ValueError):
        for product in stack:
            if product.baseline.get('insarBaseline', False):
                product.baseline = {}
            else:
                product.baseline['stateVectors']['positions'] = {}
                product.baseline['stateVectors']['velocities'] = {}
        reference.baseline = {}
        get_baseline_from_stack(reference=reference, stack=stack)


def run_test_valid_state_vectors(reference, output):
    if reference is not None:
        product = as_ASFProduct(reference, ASFSession())
        clear_baseline(reference, product)
        assert output == product.is_valid_reference()
        return


def clear_baseline(resource, product: ASFProduct):
    # Baseline values can be restored from UMM in asfProduct constructor,
    # this erases them again if the resource omitted them from the product
    if stateVectors := resource['baseline'].get('stateVectors'):
        if stateVectors.get('positions') == {}:
            product.baseline = {'stateVectors': {'positions': {}, 'velocities': {}}}

    return product



from asf_search import ASFSearchResults, ASFSession
from asf_search.ASFSearchOptions.ASFSearchOptions import ASFSearchOptions

import os
import json

from asf_search.search.search_generator import as_ASFProduct


def run_test_serialization(product=None, results=None, opts=ASFSearchOptions()):
    if product is None:
        to_serialize = ASFSearchResults([json_to_product(prod) for prod in results])
    else:
        to_serialize = ASFSearchResults([json_to_product(product)])

    with open('serialized_product.json', 'w') as f:
        f.write(json.dumps({'results': to_serialize.geojson(), 'opts': dict(opts)}))
        f.close()

    with open('serialized_product.json', 'r') as f:
        deserialized = json.loads(f.read())
        f.close()

    os.remove('serialized_product.json')

    deserialized_results = deserialized.get('results')
    deserialized_opts = deserialized.get('opts')

    for key, value in deserialized_opts.items():
        assert value == getattr(opts, key)

    for idx, original in enumerate(to_serialize):
        assert deserialized_results['features'][idx]['properties'] == original.properties
        assert deserialized_results['features'][idx]['geometry'] == original.geometry

    assert deserialized_results['type'] == 'FeatureCollection'


def json_to_product(product):
    output = as_ASFProduct(product, session=ASFSession())
    output.meta = product['meta']
    output.properties = product['properties']
    output.geometry = product['geometry']
    output.umm = product['umm']
    return output
